#ifndef OCL_UTIL_H
#define OCL_UTIL_H

#include <stdio.h>
#include <CL/cl.h>

#define KRED  "\x1B[31m"
#define KGRN  "\x1B[32m"
#define KBLK "\x1B[0m"


void printf_red(char* string) { printf("%s%s%s",KRED, string, KBLK); }
void printf_green(char* string) { printf("%s%s%s",KGRN, string, KBLK); }


static void handleError(cl_int err, const char *file, int line) 
{
    static const char* stringError[] =
    {
        // Error Codes
        "CL_SUCCESS"                                    //   0
        , "CL_DEVICE_NOT_FOUND"                         //  -1
        , "CL_DEVICE_NOT_AVAILABLE"                     //  -2
        , "CL_COMPILER_NOT_AVAILABLE"                   //  -3
        , "CL_MEM_OBJECT_ALLOCATION_FAILURE"            //  -4
        , "CL_OUT_OF_RESOURCES"                         //  -5
        , "CL_OUT_OF_HOST_MEMORY"                       //  -6
        , "CL_PROFILING_INFO_NOT_AVAILABLE"             //  -7
        , "CL_MEM_COPY_OVERLAP"                         //  -8
        , "CL_IMAGE_FORMAT_MISMATCH"                    //  -9
        , "CL_IMAGE_FORMAT_NOT_SUPPORTED"               //  -10
        , "CL_BUILD_PROGRAM_FAILURE"                    //  -11
        , "CL_MAP_FAILURE"                              //  -12
        , ""    //  -13
        , ""    //  -14
        , ""    //  -15
        , ""    //  -16
        , ""    //  -17
        , ""    //  -18
        , ""    //  -19
        , ""    //  -20
        , ""    //  -21
        , ""    //  -22
        , ""    //  -23
        , ""    //  -24
        , ""    //  -25
        , ""    //  -26
        , ""    //  -27
        , ""    //  -28
        , ""    //  -29
        , "CL_INVALID_VALUE"                            //  -30
        , "CL_INVALID_DEVICE_TYPE"                      //  -31
        , "CL_INVALID_PLATFORM"                         //  -32
        , "CL_INVALID_DEVICE"                           //  -33
        , "CL_INVALID_CONTEXT"                          //  -34
        , "CL_INVALID_QUEUE_PROPERTIES"                 //  -35
        , "CL_INVALID_COMMAND_QUEUE"                    //  -36
        , "CL_INVALID_HOST_PTR"                         //  -37
        , "CL_INVALID_MEM_OBJECT"                       //  -38
        , "CL_INVALID_IMAGE_FORMAT_DESCRIPTOR"          //  -39
        , "CL_INVALID_IMAGE_SIZE"                       //  -40
        , "CL_INVALID_SAMPLER"                          //  -41
        , "CL_INVALID_BINARY"                           //  -42
        , "CL_INVALID_BUILD_OPTIONS"                    //  -43
        , "CL_INVALID_PROGRAM"                          //  -44
        , "CL_INVALID_PROGRAM_EXECUTABLE"               //  -45
        , "CL_INVALID_KERNEL_NAME"                      //  -46
        , "CL_INVALID_KERNEL_DEFINITION"                //  -47
        , "CL_INVALID_KERNEL"                           //  -48
        , "CL_INVALID_ARG_INDEX"                        //  -49
        , "CL_INVALID_ARG_VALUE"                        //  -50
        , "CL_INVALID_ARG_SIZE"                         //  -51
        , "CL_INVALID_KERNEL_ARGS"                      //  -52
        , "CL_INVALID_WORK_DIMENSION"                   //  -53
        , "CL_INVALID_WORK_GROUP_SIZE"                  //  -54
        , "CL_INVALID_WORK_ITEM_SIZE"                   //  -55
        , "CL_INVALID_GLOBAL_OFFSET"                    //  -56
        , "CL_INVALID_EVENT_WAIT_LIST"                  //  -57
        , "CL_INVALID_EVENT"                            //  -58
        , "CL_INVALID_OPERATION"                        //  -59
        , "CL_INVALID_GL_OBJECT"                        //  -60
        , "CL_INVALID_BUFFER_SIZE"                      //  -61
        , "CL_INVALID_MIP_LEVEL"                        //  -62
        , "CL_INVALID_GLOBAL_WORK_SIZE"                 //  -63
    };

    if (err != CL_SUCCESS)
    {
       if (-err >= 0 && -err <= 63) {
          printf("error: %s in %s at line %d\n",stringError[-err],file,line);

       } else {
          printf("error: %d in %s at line %d\n",err,file,line);
       }

       //in this special case we want to get further information about the build failure (CL_BUILD_PROGRAM_FAILURE)
       if(err != -11){
          exit(-1);
       }
    }
}

#define HANDLE_ERROR( error ) (handleError( (error) , __FILE__ , __LINE__ ))

char * readFile(char* filename)
{
    // read source code of GPU kernel
    FILE* sourceFile = NULL;
    sourceFile = fopen(filename, "r");

    if(sourceFile == NULL) { printf("cannot open the kernel file\n"); exit(-1); }

    //determine file size
    fseek(sourceFile, 0, SEEK_END);
    long fileSize;
    fileSize = ftell(sourceFile);
    rewind(sourceFile);

    //allocate memory
    char *buffer = 0;
    buffer = (char*)malloc((fileSize + 1 )* sizeof(char));
    if(buffer == 0) { printf("Error: can not allocate memory\n"); exit(-1); }

    //read the file
    if(!fread(buffer, 1, fileSize, sourceFile)) { printf("Error: Can not read the file.\n"); exit(-1); }
    
    //null-terminate
    buffer[fileSize] = 0;

    fclose(sourceFile);
    return buffer;
}

typedef struct 
{
   //available platforms
   cl_uint num_platforms;
   cl_platform_id* platforms;

	cl_context context;
	cl_program program;
	cl_command_queue* command_queue;

   //available kernels
	cl_kernel saxpy;

   //available compute devices
   cl_uint num_devices;
   cl_device_id* devices;

} ocl_context_t;

/**
 * Initializes the OpenCL context
 *
 * \param filename filename of the .cl file which includes all kernels
 */
void ocl_context_init(ocl_context_t* ocl_context, char* filename)
{

   printf("---------------------------- OpenCL Initialization -------------------------\n");
   cl_int error;

   /*****************************************
    * 1) Get Platform info
    *****************************************/

   //get number of platforms available
	HANDLE_ERROR( clGetPlatformIDs(NULL, NULL, &(ocl_context->num_platforms)) );

   //get the id of the all available platforms
	ocl_context->platforms = (cl_platform_id*) malloc(sizeof(cl_platform_id) * ocl_context->num_platforms);
	HANDLE_ERROR( clGetPlatformIDs(ocl_context->num_platforms, ocl_context->platforms, NULL) );

   for(unsigned int i=0;i < ocl_context->num_platforms; ++i){
      printf("Platform %d\n",i);

      char info[2048];
      HANDLE_ERROR( clGetPlatformInfo(ocl_context->platforms[i], CL_PLATFORM_NAME, sizeof(info), info, NULL) );
      printf("Platform name: %s\n",info);

      HANDLE_ERROR( clGetPlatformInfo(ocl_context->platforms[i], CL_PLATFORM_VERSION, sizeof(info), info, NULL) );
      printf("Platform version: %s\n",info);
   }


   /*****************************************
    * 2) Create Context 
    *****************************************/

   //choose context for GPU devices
	cl_context_properties props[3];
	props[0] = (cl_context_properties)CL_CONTEXT_PLATFORM;  // indicates that next element is platform
	props[1] = (cl_context_properties)ocl_context->platforms[1];  // specifies the desired platform //TODO: this value should be determined automatically
	props[2] = (cl_context_properties)0;   // last element must be 0
   ocl_context->context = clCreateContextFromType (props, CL_DEVICE_TYPE_GPU, NULL, NULL, NULL);


   /*****************************************
    * 3) Get Devices info
    *****************************************/
   //get number of available compute devices of requested context
   HANDLE_ERROR( clGetContextInfo (ocl_context->context, CL_CONTEXT_NUM_DEVICES, sizeof(cl_uint), &(ocl_context->num_devices), NULL) ); 
   printf("%d devices available!\n",ocl_context->num_devices);

   //get compute devices
   ocl_context->devices = (cl_device_id*) malloc(sizeof(cl_device_id) * ocl_context->num_devices);
   clGetContextInfo (ocl_context->context, CL_CONTEXT_DEVICES, sizeof(cl_device_id) * ocl_context->num_devices, ocl_context->devices, NULL);

   for(unsigned int i=0;i < ocl_context->num_devices; ++i){
      cl_char info[2048];

      HANDLE_ERROR( clGetDeviceInfo(ocl_context->devices[i], CL_DEVICE_VENDOR, sizeof(info), info, NULL) );
      printf("Dendor name: %s\n", info);

      HANDLE_ERROR( clGetDeviceInfo(ocl_context->devices[i], CL_DEVICE_NAME, sizeof(info), info, NULL) );
      printf("Device name: %s\n", info);
   }


   /*****************************************
    * 4) Create Command queue for each device
    *****************************************/
   //CL command queues can be configured such that they can execute out-of-order. This is different from CUDA!
   ocl_context->command_queue = (cl_command_queue*) malloc(sizeof(cl_command_queue) * ocl_context->num_devices);

   for(unsigned int i=0;i < ocl_context->num_devices; ++i){
	   ocl_context->command_queue[i] = clCreateCommandQueue(ocl_context->context, ocl_context->devices[i], CL_QUEUE_PROFILING_ENABLE, &error);
      HANDLE_ERROR( error );
   }


   /*****************************************
    * 5) Create ocl program
    *****************************************/
	char *source = readFile(filename); 
	ocl_context->program = clCreateProgramWithSource(ocl_context->context, 1, (const char**) &source, 0, &error); HANDLE_ERROR( error );
   free(source);


   /*****************************************
    * 6) Compile ocl program
    *****************************************/
   //if you want, you can specify some compiler options here
	HANDLE_ERROR( error = clBuildProgram(ocl_context->program, ocl_context->num_devices, ocl_context->devices, NULL, NULL, NULL) );

   //get debug information
   if( error != CL_SUCCESS){

      char debug_buffer[4096];

      HANDLE_ERROR( clGetProgramBuildInfo(ocl_context->program,
                                          ocl_context->devices[0],
                                          CL_PROGRAM_BUILD_LOG, 
                                          sizeof(debug_buffer),
                                          debug_buffer, 
                                          NULL) );

      printf("%s\n", debug_buffer);

   }


   /*****************************************
    * 7) Create ocl kernels 
    *****************************************/
	ocl_context->saxpy = clCreateKernel(ocl_context->program, "saxpy", &error);
   HANDLE_ERROR( error );

   printf("--------------------------------------------------------------------------\n");

}

void ocl_context_free(ocl_context_t* ocl_context)
{
   //Cleanup
   clReleaseKernel (ocl_context->saxpy); 
   clReleaseProgram (ocl_context->program); 
   clReleaseContext (ocl_context->context);
   for(unsigned int i=0;i < ocl_context->num_devices; ++i){
      clReleaseCommandQueue(ocl_context->command_queue[i]); 
   }
}

#endif
